#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

import functools
import new


class ProxyContainerMethodsMetaClass(type):
    def __init__(cls, name, bases, dct):
        func_names_to_proxy = set(['__setitem__', '__getitem__', '__delitem__', '__contains__',
                                   '__iter__', '__len__'])
        for attr in func_names_to_proxy:
            if not hasattr(cls, attr):
                func = functools.partial(dct['_proxyfunc_'], attr)
                setattr(cls, attr, new.instancemethod(func, None, cls))
        return super(ProxyContainerMethodsMetaClass, cls).__init__(name, bases, dct)



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

import os

from flawless.lib.data_structures import ProxyContainerMethodsMetaClass


class PrefixTree(object):
    __metaclass__ = ProxyContainerMethodsMetaClass
    _proxyfunc_ = lambda attr, self, *args, **kwargs: getattr(self.root, attr)(*args, **kwargs)

    def __init__(self, split_key_func, join_key_func, accumulator_func=None,
                 accumulator_intializer=None):
        self.split_key_func = split_key_func
        self.join_key_func = join_key_func
        self.accumulator_func = accumulator_func or (lambda x, y: x)
        self.accumulator_intializer = accumulator_intializer
        self.root = Branch(self)
        self.length = 0

    def set_accumulator(self, accumulator_func, accumulator_intializer):
        self.accumulator_func = accumulator_func
        self.accumulator_intializer = accumulator_intializer


class StringPrefixTree(PrefixTree):

    def __init__(self, accumulator_func=None, accumulator_intializer=None):
        split_key_func = lambda s: (s[0], s[1:])
        join_key_func = lambda *args: "".join(args)
        super(StringPrefixTree, self).__init__(
            split_key_func=split_key_func,
            join_key_func=join_key_func,
            accumulator_func=accumulator_func,
            accumulator_intializer=accumulator_intializer,
        )


class FilePathTree(PrefixTree):

    def __init__(self, accumulator_func=None, accumulator_intializer=None, sep=os.sep):
        split_key_func = lambda s: (s, None) if sep not in s else s.split(sep, 1)
        join_key_func = lambda *args: sep.join(*args)
        super(FilePathTree, self).__init__(
            split_key_func=split_key_func,
            join_key_func=join_key_func,
            accumulator_func=accumulator_func,
            accumulator_intializer=accumulator_intializer,
        )


class Branch(object):

    def __init__(self, trunk):
        self.trunk = trunk
        self.branches = dict()
        self.size = 0
        self.value = None
        self.is_set = False

    def __str__(self):
        retval = []
        if self.is_set:
            retval.append("(%s)" % str(self.value))

        for index, (key, subbranch) in enumerate(self.branches.items()):
            pad = "|     " if index != (len(self.branches) - 1) else "        "
            subbranch_str = "\n".join([pad + s for s in str(subbranch).split("\n")])
            retval.append("|-- " + str(key))
            retval.append(subbranch_str)
        return "\n".join(retval)

    def __setitem__(self, key, value):
        if not key:
            retval = not self.is_set
            self.value = value
            self.is_set = True
            self.size += 1
            return retval

        head, remaining = self.trunk.split_key_func(key)
        if head not in self.branches:
            self.branches[head] = Branch(self.trunk)
        retval = self.branches[head].__setitem__(remaining, value)
        self.size += int(retval)
        return retval

    def __getitem__(self, key):
        if not key:
            if self.trunk.accumulator_intializer is None:
                return self.value
            else:
                return self.trunk.accumulator_func(self.trunk.accumulator_intializer, self.value)

        head, remaining = self.trunk.split_key_func(key)
        if head not in self.branches:
            retval = self.trunk.accumulator_intializer
        else:
            retval = self.branches[head][remaining]
        return self.trunk.accumulator_func(retval, self.value)

    def __delitem__(self, key):
        head, remaining = self.trunk.split_key_func(key)
        if not remaining and head in self.branches:
            del_size = self.branches[head].size
            self.size -= del_size
            del self.branches[head]
            return del_size
        if not remaining and head not in self.branches:
            return 0
        elif remaining:
            num_deleted = self.branches[head].__delitem__(remaining)
            self.size -= num_deleted
            return num_deleted

    def __contains__(self, key):
        if not key:
            return True

        head, remaining = self.trunk.split_key_func(key)
        if head not in self.branches:
            return False
        else:
            return self.branches[head][remaining]

    def __iter__(self):
        for key in self.branches:
            if self.is_set or self.branches[key].size == 1:
                yield self.trunk.join_key_func(key)
            for sub_result in self.branches[key]:
                yield self.trunk.join_key_func(key, sub_result)

    def __len__(self):
        return self.size



# NOTE: As of the time of writing, python-daemon<=1.5.5 is incompatible with recent versions of lockfile. If you
# wish to use this daemon and are using python-daemon<=1.5.5 please use lockfile==0.8

# Disable relative imports, and remove magical current directory sys.path entry
from __future__ import absolute_import
import sys
if __name__ == "__main__":
    sys.path.pop(0)

import daemon
import daemon.pidlockfile
import lockfile
import os.path
import os

import flawless.server.server


def process_is_running(pid):
    try:
        # NOOP if exists, otherwise throws exception
        os.kill(pid, 0)
        return True
    except OSError:
        return False


def get_context(pid_file_path, error_log):
    context = daemon.DaemonContext()
    context.detach_process = True
    context.working_directory = '.'
    context.stdout = error_log
    context.stderr = error_log
    if pid_file_path:
        context.pidfile = daemon.pidlockfile.TimeoutPIDLockFile(pid_file_path, acquire_timeout=2)
    return context


def main(argv):
    # Process argv
    if len(argv) <= 1:
        print "\nUsage: python flawless_daemon.py FLAWLESS_CONFIG_PATH PID_FILE_PATH RUN_DIR"
        print "    FLAWLESS_CONFIG_PATH - The path to the flawless.cfg config you want to use"
        print "    PID_FILE_PATH - The path you want to for the PID lock file"
        print "    RUN_DIR - Directory to output run data"
        return
    flawless_cfg_path = os.path.abspath(argv[1])
    pid_file_path = os.path.abspath(argv[2])
    run_dir_path = os.path.abspath(argv[3])

    # Setup logging of output
    pid = os.getpid()
    filename = os.path.join(run_dir_path, "flawless-%d.ERROR" % pid)
    error_log = open(filename, "w+", 1)

    retry = 2
    while retry > 0:
        context = get_context(pid_file_path, error_log)
        try:
            with context:
                os.setpgid(0, os.getpid())
                flawless.server.server.serve(flawless_cfg_path)
        except lockfile.LockError:
            sys.stderr.write("Error: pid file exists %s.\n" % pid_file_path)
            current_pid = context.pidfile.read_pid()
            if process_is_running(current_pid):
                sys.stderr.write("Error: Process already running %s.\n" % pid_file_path)
                sys.exit(1)
            else:
                sys.stderr.write("Error: No process running %s. Breaking lock\n" % pid_file_path)
                context.pidfile.break_lock()
                retry -= 1
        except (KeyboardInterrupt, SystemExit):
            context.pidfile.release()
            sys.exit(0)

if __name__ == '__main__':
    main(sys.argv)



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

from flawless.client.client import set_hostports
from flawless.client.client import record_error



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

import socket
import sys

try:
    import webob
except:
    pass

import flawless.client


class FlawlessMiddleware(object):
    """Middleware records errors to the error backend"""
    def __init__(self, app):
        self.app = app
        self.hostname = socket.gethostname()

    def __call__(self, environ, start_response):
        try:
            return self.app(environ, start_response)
        except:
            type, value, tb = sys.exc_info()
            reconstructed_req = self._reconstruct_request(environ)
            flawless.client.record_error(hostname=self.hostname, tb=tb, exception_message=repr(value),
                                         additional_info=reconstructed_req)
            raise value, None, tb

    def _reconstruct_request(self, environ):
        request_str = ""
        if "webob" in globals():
            request_str = str(webob.Request(environ))[:2000]
        else:
            req_parts = []
            method = environ.get("REQUEST_METHOD", "")
            path = environ.get("PATH_INFO", "")
            path += ("?" * bool(environ.get("QUERY_STRING"))) + environ.get("QUERY_STRING", "")

            req_parts.append("%s %s %s" % (method, path, environ.get("SERVER_PROTOCOL", "")))
            req_parts.append("Host: %s" % environ.get("HTTP_HOST", ""))
            req_parts.append("Referer: %s" % environ.get("HTTP_REFERER", ""))
            req_parts.append("Cookie: %s" % environ.get("HTTP_COOKIE", ""))
            req_parts.append("Content-Length: %s" % environ.get("CONTENT_LENGTH", ""))
            req_parts.append("User-Agent: %s" % environ.get("HTTP_USER_AGENT", ""))
            request_str = "\n".join(req_parts)

        return request_str



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

import sys

import flawless.lib.config
import flawless.server.server
import flawless.server.configure_server


def show_options():
    for option in flawless.lib.config.OPTIONS:
        print option.name
        print "    default value: %s" % repr(option.default)
        print "    type: %s" % str(option.type)
        print "    description: %s\n" % option.description


def usage():
    print "Usage: flawless [start|configure|options|help] [-conf path]"
    print "    Commands:"
    print "        start - Start the server"
    print "        configure - Run server setup script"
    print "        options - Display list of server configuration options for flawless.cfg"
    print "        help - Display this description"
    print "    Options:"
    print "        -conf path: Path to flawless.cfg"


def main():
    conf_path = flawless.lib.config.default_path

    command = None if len(sys.argv) > 1 else usage
    args_list = list(reversed(sys.argv[1:]))
    while args_list:
        arg = args_list.pop()
        if arg == "start" and not command:
            command = lambda: flawless.server.server.serve(conf_path)
        elif arg == "configure" and not command:
            command = lambda: flawless.server.configure_server.interview(conf_path)
        elif arg == "options" and not command:
            command = show_options
        elif arg == "help" and not command:
            command = usage
        elif arg == "-conf":
            if not args_list:
                command = usage
                break
            conf_path = args_list.pop()
        else:
            command = usage
            break

    command()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

import functools
import traceback
import linecache
import os.path
import random
import socket
import sys
import time
import warnings

from thrift.Thrift import TException
from thrift.transport import TTransport
from thrift.transport import TSocket
from thrift.protocol import TBinaryProtocol

import flawless.client.default
import flawless.lib.config
import flawless.server.api.ttypes as api_ttypes
from flawless.server.api import Flawless


config = flawless.lib.config.get()

MAX_VARIABLE_REPR = 250
MAX_LOCALS = 100
NUM_FRAMES_TO_SAVE = 20

BACKOFF_MS = 0
CONSECUTIVE_CONNECTION_ERRORS = 0


def _get_epoch_ms():
    return int(time.time() * 1000)


def _get_backend_host():
    hostports = config.flawless_hostports or flawless.client.default.hostports
    return random.choice(hostports) if hostports else None


def _get_service():
    hostport = _get_backend_host()
    if not hostport:
        warnings.warn("Unable to record error: flawless server hostport not set", RuntimeWarning)
        return None, None

    host, port = hostport.split(":")
    tsocket = TSocket.TSocket(host, int(port))
    tsocket.setTimeout(2000)  # 2 second timeout
    transport = TTransport.TFramedTransport(tsocket)
    protocol = TBinaryProtocol.TBinaryProtocol(transport)
    client = Flawless.Client(protocol)
    return client, transport


def _myrepr(s):
    try:
        repr_str = repr(s)
        return repr_str[:MAX_VARIABLE_REPR] + "..." * int(len(repr_str) > MAX_VARIABLE_REPR)
    except:
        return "Could not except repr for this field"


def set_hostports(hostports):
    if type(hostports) not in [tuple, list]:
        raise ValueError("hostports must be a list or tuple")
    flawless.client.default.hostports = hostports


def record_error(hostname, sys_traceback, exception_message, preceding_stack=None,
                 error_threshold=None, additional_info=None):
    ''' Helper function to record errors to the flawless backend '''
    stack = []
    while sys_traceback is not None:
        stack.append(sys_traceback)
        sys_traceback = sys_traceback.tb_next

    stack_lines = []
    for row in preceding_stack or []:
        stack_lines.append(
            api_ttypes.StackLine(filename=os.path.abspath(row[0]), line_number=row[1],
                                 function_name=row[2], text=row[3])
        )

    for index, tb in enumerate(stack):
        filename = tb.tb_frame.f_code.co_filename
        func_name = tb.tb_frame.f_code.co_name
        lineno = tb.tb_lineno
        line = linecache.getline(filename, lineno, tb.tb_frame.f_globals)
        frame_locals = None
        if index >= (len(stack) - NUM_FRAMES_TO_SAVE):
            # Include some limits on max string length & number of variables to keep things from getting
            # out of hand
            frame_locals = dict((k, _myrepr(v)) for k, v in
                                tb.tb_frame.f_locals.items()[:MAX_LOCALS] if k != "self")
            if "self" in tb.tb_frame.f_locals and hasattr(tb.tb_frame.f_locals["self"], "__dict__"):
                frame_locals.update(dict(("self." + k, _myrepr(v)) for k, v in
                                         tb.tb_frame.f_locals["self"].__dict__.items()[:MAX_LOCALS]
                                         if k != "self"))

        # TODO (john): May need to prepend site-packages to filename to get correct path
        stack_lines.append(
            api_ttypes.StackLine(filename=os.path.abspath(filename), line_number=lineno,
                                 function_name=func_name, text=line, frame_locals=frame_locals)
        )

    req = api_ttypes.RecordErrorRequest(
        traceback=stack_lines,
        exception_message=exception_message,
        hostname=hostname,
        error_threshold=error_threshold,
        additional_info=additional_info,
    )

    global BACKOFF_MS
    global CONSECUTIVE_CONNECTION_ERRORS
    client, transport = _get_service()
    try:
        if client and transport and _get_epoch_ms() >= BACKOFF_MS:
            transport.open()
            client.record_error(req)
            CONSECUTIVE_CONNECTION_ERRORS = int(CONSECUTIVE_CONNECTION_ERRORS / 2)
            if CONSECUTIVE_CONNECTION_ERRORS > 0:
                backoff = 1000 * random.randint(1, 2 ** CONSECUTIVE_CONNECTION_ERRORS)
                BACKOFF_MS = _get_epoch_ms() + backoff

    except TException:
        CONSECUTIVE_CONNECTION_ERRORS = max(12, CONSECUTIVE_CONNECTION_ERRORS + 1)
        backoff = 1000 * random.randint(1, 2 ** CONSECUTIVE_CONNECTION_ERRORS)
        BACKOFF_MS = _get_epoch_ms() + backoff
        raise
    finally:
        if transport and transport.isOpen():
            transport.close()


def _safe_wrap(func):
    safe_attrs = [attr for attr in functools.WRAPPER_ASSIGNMENTS if hasattr(func, attr)]
    return functools.wraps(func, safe_attrs)


def _wrap_function_with_error_decorator(func,
                                        save_current_stack_trace=True,
                                        reraise_exception=True,
                                        error_threshold=None):
    preceding_stack = []
    if save_current_stack_trace:
        preceding_stack = traceback.extract_stack()

    @_safe_wrap(func)
    def wrapped_func_with_error_reporting(*args, **kwargs):
        if not _get_backend_host():
            warnings.warn("flawless server hostport not set", RuntimeWarning, stacklevel=2)
        try:
            return func(*args, **kwargs)
        except:
            type, value, sys_traceback = sys.exc_info()

            # Check to try and prevent multiple reports of the same exception
            if hasattr(value, "_flawless_already_caught"):
                if reraise_exception:
                    raise value, None, sys_traceback
                else:
                    return

            # Get trackback & report it
            hostname = socket.gethostname()
            record_error(
                hostname=hostname,
                sys_traceback=sys_traceback,
                preceding_stack=preceding_stack,
                exception_message=repr(value),
                error_threshold=error_threshold,
            )

            # Reraise exception if so desired
            if reraise_exception:
                setattr(value, "_flawless_already_caught", True)
                raise value, None, sys_traceback
    return wrapped_func_with_error_reporting



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

from flawless.lib.storage import StorageInterface


class StubStorage(StorageInterface):

    def __init__(self, partition):
        self.partition = partition
        self.dict = dict()

    def iteritems(self):
        return self.dict.iteritems()

    def __setitem__(self, key, item):
        self.dict[key] = item

    def __getitem__(self, key):
        return self.dict.get(key)

    def __contains__(self, key):
        return key in self.dict



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
import logging
import os
import os.path
from SocketServer import ThreadingMixIn
import signal
import sys
import urlparse

from thrift.protocol import TBinaryProtocol
from thrift.server import TServer
from thrift.transport import TSocket
from thrift.transport import TTransport

import flawless.lib.config
from flawless.lib.storage import DiskStorage
from flawless.server.api import Flawless
from flawless.server.service import FlawlessThriftServiceHandler
from flawless.server.service import FlawlessWebServiceHandler

log = logging.getLogger(__name__)
config = flawless.lib.config.get()


class SimpleThreadedHTTPServer(ThreadingMixIn, HTTPServer):
    def attach_service(self, service):
        self.service = service

    def server_close(self):
        HTTPServer.server_close(self)


class SimpleRequestHTTPHandler(BaseHTTPRequestHandler):

    def do_GET(self):
        parts = urlparse.urlparse(self.path)
        kwargs = dict(urlparse.parse_qsl(parts.query))
        ret = None

        try:
            if hasattr(self.server.service, parts.path[1:] or "index"):
                ret = getattr(self.server.service, parts.path[1:] or "index")(**kwargs)
                self.send_response(200)
                self.send_header('Content-Length', len(ret or ""))
                self.send_header('Content-Type', 'text/html')
            else:
                self.send_response(404)
        except Exception as e:
            log.exception(e)
            self.send_response(500)
        finally:
            self.end_headers()
        if ret:
            self.wfile.write(ret)

    def do_POST(self):
        # Read in POST body
        parts = urlparse.urlparse(self.path)
        content_length = int(self.headers.getheader("Content-Length"))
        req_str = self.rfile.read(content_length)

        ret = None
        try:
            ret = getattr(self.server.service, parts.path[1:])(req_str)
            self.send_response(200)
            if ret:
                self.send_header('Content-Length', len(ret))
                self.send_header('Content-Type', 'text/html')
        except Exception as e:
            log.exception(e)
            self.send_response(500)
        finally:
            self.end_headers()

        if ret:
            self.wfile.write(ret)


def serve(conf_path, storage_factory=None):
    """This method starts the server. There are two processes, one is an HTTP server that shows
    and admin interface and the second is a Thrift server that the client code calls.

    Arguments:
        `conf_path` - The path to your flawless.cfg file
        `storage_factory` - You can pass in your own storage class that implements StorageInterface. You must implement
                            storage_cls if you want Flawless to be horizontally scalable, since by default it will just
                            store everything on the local disk.
    """

    flawless.lib.config.init_config(conf_path)
    # Try and create datadir if it doesn't exist. For instance it might be in /tmp
    if not os.path.exists(config.data_dir_path):
        os.makedirs(config.data_dir_path)

    storage_factory = storage_factory or (lambda partition: DiskStorage(partition=partition))
    logging.basicConfig(level=getattr(logging, config.log_level), filename=config.log_file, stream=sys.stderr)
    child_pid = os.fork()
    if child_pid == 0:
        # Setup HTTP server
        handler = FlawlessWebServiceHandler(storage_factory=storage_factory)
        server = SimpleThreadedHTTPServer(('', config.http_port), SimpleRequestHTTPHandler)
        server.attach_service(handler)
        server.request_queue_size = 50

        try:
            server.serve_forever()
        except (KeyboardInterrupt, SystemExit):
            server.server_close()
    else:
        # Setup Thrift server
        handler = FlawlessThriftServiceHandler(storage_factory=storage_factory)
        processor = Flawless.Processor(handler)
        transport = TSocket.TServerSocket(port=config.port)
        tfactory = TTransport.TFramedTransportFactory()
        pfactory = TBinaryProtocol.TBinaryProtocolFactory()
        server = TServer.TThreadedServer(processor, transport, tfactory, pfactory)
        try:
            server.serve()
        except (KeyboardInterrupt, SystemExit):
            handler.errors_seen.sync()
            transport.close()
            os.kill(child_pid, signal.SIGINT)


def main():
    conf_path = flawless.lib.config.default_path
    if len(sys.argv) > 1:
        conf_path = sys.argv[1]
    serve(conf_path)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

from flawless.server.api import Flawless


class FlawlessServiceStub(Flawless.Iface):

    def __init__(self):
        for func in [f for f in dir(self) if not f.startswith("_")]:
            getattr(self, func).__dict__["result"] = None
            getattr(self, func).__dict__["last_args"] = None
            getattr(self, func).__dict__["args_list"] = list()

    def _handle_stub(self, func, args):
        last_args = dict((k, v) for k, v in args.items() if k != "self")
        getattr(self, func).__dict__["last_args"] = last_args
        getattr(self, func).__dict__["args_list"].append(last_args)
        return getattr(self, func).__dict__["result"]

    def record_error(self, request):
        return self._handle_stub("record_error", locals())



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

import abc
import logging
import os
import os.path
import re
import subprocess

import flawless.lib.config


log = logging.getLogger(__name__)
config = flawless.lib.config.get()


def get_repository(open_process_func=subprocess.Popen):
    if config.repo_type == "git":
        return GitRepository(open_process_func=open_process_func)


class Repository(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, local_path=None, remote_url=None, branch_pattern=None,
                 open_process_func=subprocess.Popen):
        self.local_path = local_path or config.repo_dir
        self.remote_url = remote_url or config.repo_url
        self.open_process_func = open_process_func
        self.branch_pattern = re.compile(config.repo_branch_pattern) if config.repo_branch_pattern else None

    @abc.abstractmethod
    def blame(self, filename, line_number):
        pass

    @abc.abstractmethod
    def update(self):
        pass

    @abc.abstractmethod
    def create(self):
        pass


class GitRepository(Repository):

    def __init__(self, *args, **kwargs):
        super(GitRepository, self).__init__(*args, **kwargs)
        self.extract_email_pattern = re.compile(r"^author-mail <([^>]+)+>$")
        self.extract_modified_pattern = re.compile(r"^author-time (\d+)$")
        self.digit_tokenizer_pattern = re.compile(r'(\d+)|(\D+)').findall
        self.natural_sort_func = \
            lambda s: tuple(int(num) if num else alpha for num, alpha in self.digit_tokenizer_pattern(s))

    def _raw_run(self, args, log_output=False):
        p = self.open_process_func(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        err = p.stderr.read()
        out = p.stdout.read()
        if log_output and out:
            log.info(out)
        if log_output and err:
            log.error(err)
        return out

    def _run_git_command(self, args, log_output=False):
        base_args = [
            config.git_cli_path,
            "--git-dir=%s" % os.path.join(self.local_path, ".git"),
            "--work-tree=%s" % self.local_path,
        ]
        base_args.extend(args)
        return self._raw_run(base_args, log_output=log_output)

    def blame(self, filename, line_number):
        args = [
            "blame",
            "-p",
            os.path.join(self.local_path, filename),
            "-L",
            "%d,+1" % line_number,
        ]
        output = self._run_git_command(args)
        email, modified = None, None
        for line in output.split("\n"):
            if email and modified:
                break

            match = self.extract_email_pattern.match(line)
            if match:
                email = match.group(1)
            match = self.extract_modified_pattern.match(line)
            if match:
                modified = int(match.group(1))

        return email, modified

    def update(self, log_output=False):
        if not self.remote_url:
            return

        if self.branch_pattern:
            branch_names = self._run_git_command(["fetch"], log_output=log_output)
            all_branches = self._run_git_command(["branch", "-r"], log_output=log_output).split("\n")
            all_branches = [s.strip() for s in all_branches if "->" not in s and s.strip()]
            all_branches = sorted(all_branches, key=self.natural_sort_func)
            all_branches = [s for s in all_branches if self.branch_pattern.match(s)]
            self._run_git_command(["reset", "--hard", all_branches[-1]], log_output=log_output)
        else:
            self._run_git_command(["pull", "--rebase"], log_output=log_output)

    def create(self):
        if not os.path.exists(self.local_path):
            os.makedirs(self.local_path)
        self._raw_run(
            [config.git_cli_path, "clone", self.remote_url, self.local_path],
            log_output=True,
        )
        self.update(log_output=True)



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

import ConfigParser
import os
import os.path
import subprocess
import socket
import shutil

import flawless.lib.config
from flawless.lib.version_control import repo


def interview(conf_path):
    options = dict(
        log_level="WARNING",
        port="9028",
    )

    print "Configure server host & port"
    print "----------------------------"
    options["port"] = raw_input("What port should the server listen on (suggested: 9028): ")
    options["hostname"] = raw_input("Internet browser accessible hostname to access this server "
                                    "(ex: http://%s): " % socket.gethostname())

    print "\nConfigure Email"
    print "---------------"
    options["smtp_host"] = raw_input("What is host:port for your smtp server (ex: smtphost:25): ").strip()
    smtp_user = raw_input("What is the username should flawless use to access your smtp "
                          "server (leave blank if username not required)? ").strip()
    if smtp_user:
        options["smtp_user"] = smtp_user
        options["smtp_password"] = raw_input("What is %s's password? " % smtp_user).strip()
    options["email_domain_name"] = raw_input("What is your email domain (ex: example.com): ").strip()
    options["ignore_vcs_email_domain"] = raw_input("Are all developer emails on %s (y/n): " %
                                                   options["email_domain_name"])[0] in ['y', 'Y']
    options["default_contact"] = raw_input("If Flawless can't figure out which developer to email, "
                                           "what should be the default email address that Flawless "
                                           "sends to? ").strip()

    print "\nConfigure Directory Paths"
    print "-------------------------"
    conf_dir_path = os.path.dirname(conf_path)
    if (raw_input("Use %s as the directory to store configuration information (y/n)? " % conf_dir_path).strip()[0]
            in ['n', 'N']):
        conf_dir_path = raw_input("Enter desired configuration directory path: ").strip()
        if not os.path.exists(conf_dir_path):
            os.makedirs(conf_dir_path)
            copy_files = True
        else:
            copy_files = raw_input("Overwrite existing files in %s (y/n)? " % conf_dir_path).strip()[0] in ['y', 'Y']

        if copy_files:
            default_dir = os.path.dirname(flawless.lib.config.default_path)
            files_to_copy = [f for f in os.listdir(default_dir)
                             if not f.startswith('.')]
            for filename in files_to_copy:
                shutil.copy2(os.path.join(default_dir, filename), conf_dir_path)
    options["data_dir_path"] = raw_input("Enter path to directory were Flawless can persist "
                                         "error data to disk (ex: /tmp/flawless/): ").strip()

    print "\nConfigure Repository Access"
    print "---------------------------"

    if subprocess.call(["which", "git"], stdout=open(os.devnull)):
        print "Could not detect git. Please enter path to git executable"
        git_cli_path = raw_input("path to git binary: ")
        options["git_cli_path"] = git_cli_path

    print "Flawless needs access to a regularly updated copy of your repo in order to run git-blame"
    print "and determine which developer caused an error. Flawless can either checkout"
    print "and manage the repo iteself, or you can just provide the directory path to"
    print "the repo and mange keeping the repo up to date yourself\n"
    repo_url = repo_dir = branch_pattern = None
    if raw_input("Should Flawless checkout a copy of your repo (y/n): ")[0] in ['y', 'Y']:
        print "Please enter the URI for the repo, including username & password if necessary"
        print "Examples: https://username:password@example.com/path/repo.git or"
        print "git://username@example.com/repo.git"
        repo_url = raw_input("Git remote URI: ").strip()
        repo_dir = raw_input("Where should the repo live (ex: /tmp/flawless_repo): ").strip()

        if raw_input("Do you cut release branches for the code flawless will monitor (y/n)? ")[0] in ['y', 'Y']:
            print "Flawless will regularly check for new branches and always work off the latest branch"
            print "Latest branch is determined by sorting branch names by name"
            print "Please enter a Python regular expression that identifies your release branch names"
            print "Example: example-server-release"
            branch_pattern = raw_input("regexp: ").strip()
    else:
        repo_dir = raw_input("What is the filepath to your git repo (ex: /tmp/repo): ").strip()

    options["repo_dir"] = repo_dir
    options["repo_url"] = repo_url
    options["repo_branch_pattern"] = branch_pattern
    options["repo_type"] = "git"

    print "\nBasic configuration done"
    if raw_input("Overwrite flawless.cfg (y/n)? ")[0] in ['y', 'Y']:
        parser = ConfigParser.SafeConfigParser()
        config_path = os.path.join(conf_dir_path, "flawless.cfg")
        parser.read(config_path)
        if not parser.has_section("flawless"):
            parser.add_section("flawless")
        for option, value in options.items():
            if value is not None:
                parser.set("flawless", str(option), str(value))
        with open(config_path, "w") as fh:
            parser.write(fh)

    if repo_url:
        print "\nYou elected to have flawless manage a copy of your git repo. Flawless will now attempt to"
        print "clone your repo"
        new_repo = repo.GitRepository(local_path=repo_dir,
                                      remote_url=repo_url,
                                      branch_pattern=branch_pattern)
        new_repo.create()

    print "\n\nSetup complete!"
    print "Start the server by running:"
    print "flawless start -conf %s" % os.path.join(conf_dir_path, "flawless.cfg")
    print ("Check server is running by visiting http://%s:%s/check_health" %
           (options["hostname"], options["port"]))


if __name__ == '__main__':
    interview(flawless.lib.config.default_path)



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

import flawless.client
import flawless.client.decorators
import flawless.lib.config


# You can wrap a function using the flawless decorator and any exceptions that
# get thrown will be reported to the flawless backend & then re-raised
@flawless.client.decorators.wrap_function
def example1():
  raise Exception()



# You can also control behavior of the decorator. For instance you can set the
# number of times an error must occur before an email gets sent. You can also
# prevent the exception from being re-raised
@flawless.client.decorators.wrap_function(error_threshold=1, reraise_exception=False)
def example2():
  raise Exception()



# Finally, you can decorate an entire class. The class decorator wraps any instance
# method or classmethod in the class with the function decorator.
@flawless.client.decorators.wrap_class
class ExampleClass(object):

  def func1(self):
    raise Exception()

  @classmethod
  def func2(cls):
    raise Exception()



if __name__ == '__main__':
  # The client has three options to configure the flawless client
  # Option 1: Set flawless_hostport in the config file and call flawless.lib.config.init_config
  flawless.lib.config.init_config("../config/flawless.cfg")

  # Option 2: Manually edit flawless/client/default.py and hardcode the value for hostport

  # Option 3: Call set_hostport
  flawless.client.set_hostport("localhost:9028")



  # example1 will re-raise the exception
  try:
    example1()
  except:
    pass

  # example 2 will not
  example2()

  # The class methods will re-raise the exception
  obj = ExampleClass()
  try:
    obj.func1()
  except:
    pass



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

import os

import flawless.lib.config
from flawless.lib.data_structures.persistent_dictionary import PersistentDictionary
from flawless.lib.storage import StorageInterface


class DiskStorage(StorageInterface):

    def __init__(self, partition):
        super(DiskStorage, self).__init__(partition)
        config = flawless.lib.config.get()
        if self.partition:
            filepath = os.path.join(config.data_dir_path, "flawless-errors-", partition)
        else:
            filepath = os.path.join(config.data_dir_path, "flawless-whitelists-config")
        self.disk_dict = PersistentDictionary(filepath)

    def _proxyfunc_(attr, self, *args, **kwargs):
        try:
            return getattr(self.disk_dict, attr)(*args, **kwargs)
        except KeyError:
            return None

    def open(self):
        self.disk_dict.open()

        # Build new copy of dict since migrate_thrift_obj may change the hash code of the keys of the dict
        migrated_dict = dict()
        for key, value in self.disk_dict.dict.items():
            self.migrate_thrift_obj(key)
            self.migrate_thrift_obj(value)
            migrated_dict[key] = value

        self.disk_dict.dict = migrated_dict

    def sync(self):
        self.disk_dict.sync()

    def close(self):
        self.disk_dict.close()

    def iteritems(self):
        return self.disk_dict.dict.iteritems()

    def __setitem__(self, key, item):
        self.disk_dict[key] = item

    def __getitem__(self, key):
        try:
            return self.disk_dict[key]
        except KeyError:
            return None

    def __contains__(self, key):
        return key in self.disk_dict



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

import collections
import ConfigParser
import os.path

FlawlessOption = collections.namedtuple("FlawlessOption", ["name", "default", "type", "description"])

OPTIONS = [
    # Hostnames & ports
    FlawlessOption("port", 9028, int, "The port to listen on"),
    FlawlessOption("http_port", 8080, int, "The port to listen on for HTTP requests"),
    FlawlessOption("hostname", "localhost", str,
                   "The host:port to connect to for the flawless backend service. This is used for "
                   "generating URLs in the emails that are sent out"),

    # Email domain name
    FlawlessOption("email_domain_name", "example.com", str, "Domain name used to send emails"),
    FlawlessOption("ignore_vcs_email_domain", False, bool,
                   "Version control systems sometimes report bogus email domains, set this if"
                   "the email domain from the version control system should be ignored and if"
                   "email_domain_name should be used instead"),
    FlawlessOption("default_contact", None, str,
                   "The default email address to contact for issues when Flawless is unable to "
                   "determine who to email"),

    # SMTP host setup
    FlawlessOption("smtp_host", "localhost:25", str, "The host responsbile for sending email"),
    FlawlessOption("smtp_user", None, str, "Username for smtp server (if required)"),
    FlawlessOption("smtp_password", None, str, "Password for smtp server (if required)"),

    # Directory paths
    FlawlessOption("data_dir_path", "/tmp/", str, "Path to where flawless can store state to disk"),
    FlawlessOption("config_dir_path", None, str, "Path to where flawless config files can be found"),

    # Control when errors get reported
    FlawlessOption("report_runtime_package_directory_names", "site-packages",
                   lambda s: [e.strip() for e in s.split(",") if e.strip()],
                   "The site-packages directory name"),
    FlawlessOption("report_only_after_minimum_date", "1970-1-1", str,
                   "Ignore errors occurring in old code that was last modified prior to this date"),
    FlawlessOption("report_error_threshold", 3, int,
                   "Number of times an error should occur before generating an alert"),
    FlawlessOption("only_blame_filepaths_matching", r"^(?!.*\.egg).*$",
                   lambda s: [e.strip() for e in s.split(",") if e.strip()],
                   "When assigning blame to a file, you can explicity only blame "
                   "files whose filepath matches any pattern on this CSV list"),
    FlawlessOption("max_concurrent_git_blames", 10, int,
                   "The maximum number of git blames that can run at the same time"),

    # Repository info
    FlawlessOption("git_cli_path", "git", str, "Path to the git command line utility"),
    FlawlessOption("repo_type", "git", str, "Options: git"),
    FlawlessOption("repo_dir", "/tmp/flawless_repo", str, "The path to the repo directory"),
    FlawlessOption("repo_url", None, str, "URL of the remote repo, including username and "
                   "password credentials"),
    FlawlessOption("repo_branch_pattern", None, str, "Regular expressions to identify release branches"),

    # Client settings
    FlawlessOption("flawless_hostports", None,
                   lambda s: [e.strip() for e in s.split(",") if e.strip()],
                   "Host:port of the flawless backend"),
    FlawlessOption("client_timeout", 10, int, "The number of seconds before requests should timeout"),

    # Logging
    FlawlessOption("log_level", "INFO", str, "The logging level"),
    FlawlessOption("log_file", None, str, "Path to the log file"),

    # Storage
    FlawlessOption("redis_version", "2.0", str, "The version of the redis server if using redis"),
]


class FlawlessConfig(object):
    def __init__(self, dict):
        self.__dict__ = dict

    def __str__(self):
        return str(self.__dict__)


RUNTIME_CONFIG = FlawlessConfig(
    dict((o.name, o.type(o.default) if o.default is not None else None) for o in OPTIONS)
)


default_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../../config/flawless.cfg"))


def get():
    return RUNTIME_CONFIG


def init_config(filepath):
    setattr(RUNTIME_CONFIG, "config_dir_path", os.path.dirname(filepath))
    parser = ConfigParser.SafeConfigParser()
    parser.read(filepath)
    for option in OPTIONS:
        if parser.has_option("flawless", option.name):
            value = parser.get("flawless", option.name)
            setattr(RUNTIME_CONFIG, option.name, option.type(value))



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

import abc
import copy


class StorageInterface(object):
    """By default Flawless stores everything on disk which means there can only be one centralized instance of
    Flawless. You can implement your own instance of StorageInterface that connects to a backend database and pass
    it into flawless.server.server.serve. Then it is possible to have the flawless server be horizontally scalable
    since the database serves as the centralized source of truth.

    It is worth noting is that the keys used in this interface are either python primitives (string, list, etc) or
    thrift objects. The values can also be python primitives (list, dictionary, etc) or thrift objects
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, partition):
        """partition is a string used to partition keys by week. For instance, with disk storage, we create a new
        file for every unique partition. For a database you may want to consider prepending all keys with
        partition. For accessing config data, the partition is None"""
        self.partition = partition

    def open(self):
        """Called to create connection to storage"""
        pass

    def sync(self):
        """Called periodically to flush data"""
        pass

    def close(self):
        """Called to close connection to storage"""
        pass

    def migrate_thrift_obj(self, obj):
        """Helper function that can be called when serializing/deserializing thrift objects whose definitions
        have changed, we need to make sure we initialize the new attributes to their default value"""
        if not hasattr(obj, "thrift_spec"):
            return

        obj_key_set = set(obj.__dict__.keys())
        thrift_field_map = {t[2]: t[4] for t in obj.thrift_spec if t}
        obj.__dict__.update({f: copy.copy(thrift_field_map[f]) for f in set(thrift_field_map.keys()) - obj_key_set})
        for value in obj.__dict__.itervalues():
            self.migrate_thrift_obj(value)

    @abc.abstractmethod
    def iteritems(self):
        """Should return iterator of tuples (key, value) for all entries for the given self.partition"""
        pass

    @abc.abstractmethod
    def __setitem__(self, key, item):
        pass

    @abc.abstractmethod
    def __getitem__(self, key):
        """Should return None if key does not exist"""
        pass

    @abc.abstractmethod
    def __contains__(self, key):
        pass



#
# Autogenerated by Thrift Compiler (0.8.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:new_style
#

from thrift.Thrift import TType, TMessageType, TException

from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TProtocol
try:
  from thrift.protocol import fastbinary
except:
  fastbinary = None


class LineType(object):
  """
  Author: jwegan@gmail.com (John Egan)
  """
  DEFAULT = 1
  KNOWN_ERROR = 2
  BUILDING_BLOCK = 3
  THIRDPARTY_WHITELIST = 4
  IGNORED_FILEPATH = 5
  BAD_FILEPATH = 6
  RAISED_EXCEPTION = 7

  _VALUES_TO_NAMES = {
    1: "DEFAULT",
    2: "KNOWN_ERROR",
    3: "BUILDING_BLOCK",
    4: "THIRDPARTY_WHITELIST",
    5: "IGNORED_FILEPATH",
    6: "BAD_FILEPATH",
    7: "RAISED_EXCEPTION",
  }

  _NAMES_TO_VALUES = {
    "DEFAULT": 1,
    "KNOWN_ERROR": 2,
    "BUILDING_BLOCK": 3,
    "THIRDPARTY_WHITELIST": 4,
    "IGNORED_FILEPATH": 5,
    "BAD_FILEPATH": 6,
    "RAISED_EXCEPTION": 7,
  }


class ErrorKey(object):
  """
  Attributes:
   - filename
   - line_number
   - function_name
   - text
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'filename', None, None, ), # 1
    (2, TType.I64, 'line_number', None, None, ), # 2
    (3, TType.STRING, 'function_name', None, None, ), # 3
    (4, TType.STRING, 'text', None, None, ), # 4
  )

  def __init__(self, filename=None, line_number=None, function_name=None, text=None,):
    self.filename = filename
    self.line_number = line_number
    self.function_name = function_name
    self.text = text

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.filename = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I64:
          self.line_number = iprot.readI64();
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.STRING:
          self.function_name = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.STRING:
          self.text = iprot.readString();
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('ErrorKey')
    if self.filename is not None:
      oprot.writeFieldBegin('filename', TType.STRING, 1)
      oprot.writeString(self.filename)
      oprot.writeFieldEnd()
    if self.line_number is not None:
      oprot.writeFieldBegin('line_number', TType.I64, 2)
      oprot.writeI64(self.line_number)
      oprot.writeFieldEnd()
    if self.function_name is not None:
      oprot.writeFieldBegin('function_name', TType.STRING, 3)
      oprot.writeString(self.function_name)
      oprot.writeFieldEnd()
    if self.text is not None:
      oprot.writeFieldBegin('text', TType.STRING, 4)
      oprot.writeString(self.text)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class StackLine(object):
  """
  Attributes:
   - filename
   - line_number
   - function_name
   - text
   - frame_locals
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'filename', None, None, ), # 1
    (2, TType.I64, 'line_number', None, None, ), # 2
    (3, TType.STRING, 'function_name', None, None, ), # 3
    (4, TType.STRING, 'text', None, None, ), # 4
    (5, TType.MAP, 'frame_locals', (TType.STRING,None,TType.STRING,None), None, ), # 5
  )

  def __init__(self, filename=None, line_number=None, function_name=None, text=None, frame_locals=None,):
    self.filename = filename
    self.line_number = line_number
    self.function_name = function_name
    self.text = text
    self.frame_locals = frame_locals

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.filename = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I64:
          self.line_number = iprot.readI64();
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.STRING:
          self.function_name = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.STRING:
          self.text = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 5:
        if ftype == TType.MAP:
          self.frame_locals = {}
          (_ktype1, _vtype2, _size0 ) = iprot.readMapBegin() 
          for _i4 in xrange(_size0):
            _key5 = iprot.readString();
            _val6 = iprot.readString();
            self.frame_locals[_key5] = _val6
          iprot.readMapEnd()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('StackLine')
    if self.filename is not None:
      oprot.writeFieldBegin('filename', TType.STRING, 1)
      oprot.writeString(self.filename)
      oprot.writeFieldEnd()
    if self.line_number is not None:
      oprot.writeFieldBegin('line_number', TType.I64, 2)
      oprot.writeI64(self.line_number)
      oprot.writeFieldEnd()
    if self.function_name is not None:
      oprot.writeFieldBegin('function_name', TType.STRING, 3)
      oprot.writeString(self.function_name)
      oprot.writeFieldEnd()
    if self.text is not None:
      oprot.writeFieldBegin('text', TType.STRING, 4)
      oprot.writeString(self.text)
      oprot.writeFieldEnd()
    if self.frame_locals is not None:
      oprot.writeFieldBegin('frame_locals', TType.MAP, 5)
      oprot.writeMapBegin(TType.STRING, TType.STRING, len(self.frame_locals))
      for kiter7,viter8 in self.frame_locals.items():
        oprot.writeString(kiter7)
        oprot.writeString(viter8)
      oprot.writeMapEnd()
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class RecordErrorRequest(object):
  """
  Attributes:
   - traceback
   - exception_message
   - hostname
   - error_threshold
   - additional_info
  """

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'traceback', (TType.STRUCT,(StackLine, StackLine.thrift_spec)), None, ), # 1
    (2, TType.STRING, 'exception_message', None, None, ), # 2
    (3, TType.STRING, 'hostname', None, None, ), # 3
    (4, TType.I64, 'error_threshold', None, None, ), # 4
    (5, TType.STRING, 'additional_info', None, None, ), # 5
  )

  def __init__(self, traceback=None, exception_message=None, hostname=None, error_threshold=None, additional_info=None,):
    self.traceback = traceback
    self.exception_message = exception_message
    self.hostname = hostname
    self.error_threshold = error_threshold
    self.additional_info = additional_info

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.LIST:
          self.traceback = []
          (_etype12, _size9) = iprot.readListBegin()
          for _i13 in xrange(_size9):
            _elem14 = StackLine()
            _elem14.read(iprot)
            self.traceback.append(_elem14)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.exception_message = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.STRING:
          self.hostname = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.I64:
          self.error_threshold = iprot.readI64();
        else:
          iprot.skip(ftype)
      elif fid == 5:
        if ftype == TType.STRING:
          self.additional_info = iprot.readString();
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('RecordErrorRequest')
    if self.traceback is not None:
      oprot.writeFieldBegin('traceback', TType.LIST, 1)
      oprot.writeListBegin(TType.STRUCT, len(self.traceback))
      for iter15 in self.traceback:
        iter15.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.exception_message is not None:
      oprot.writeFieldBegin('exception_message', TType.STRING, 2)
      oprot.writeString(self.exception_message)
      oprot.writeFieldEnd()
    if self.hostname is not None:
      oprot.writeFieldBegin('hostname', TType.STRING, 3)
      oprot.writeString(self.hostname)
      oprot.writeFieldEnd()
    if self.error_threshold is not None:
      oprot.writeFieldBegin('error_threshold', TType.I64, 4)
      oprot.writeI64(self.error_threshold)
      oprot.writeFieldEnd()
    if self.additional_info is not None:
      oprot.writeFieldBegin('additional_info', TType.STRING, 5)
      oprot.writeString(self.additional_info)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class ErrorInfo(object):
  """
  Attributes:
   - error_count
   - developer_email
   - date
   - email_sent
   - last_occurrence
   - is_known_error
   - last_error_data
  """

  thrift_spec = (
    None, # 0
    (1, TType.I64, 'error_count', None, None, ), # 1
    (2, TType.STRING, 'developer_email', None, None, ), # 2
    (3, TType.STRING, 'date', None, None, ), # 3
    (4, TType.BOOL, 'email_sent', None, None, ), # 4
    (5, TType.STRING, 'last_occurrence', None, None, ), # 5
    (6, TType.BOOL, 'is_known_error', None, None, ), # 6
    (7, TType.STRUCT, 'last_error_data', (RecordErrorRequest, RecordErrorRequest.thrift_spec), None, ), # 7
  )

  def __init__(self, error_count=None, developer_email=None, date=None, email_sent=None, last_occurrence=None, is_known_error=None, last_error_data=None,):
    self.error_count = error_count
    self.developer_email = developer_email
    self.date = date
    self.email_sent = email_sent
    self.last_occurrence = last_occurrence
    self.is_known_error = is_known_error
    self.last_error_data = last_error_data

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.I64:
          self.error_count = iprot.readI64();
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.developer_email = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.STRING:
          self.date = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.BOOL:
          self.email_sent = iprot.readBool();
        else:
          iprot.skip(ftype)
      elif fid == 5:
        if ftype == TType.STRING:
          self.last_occurrence = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 6:
        if ftype == TType.BOOL:
          self.is_known_error = iprot.readBool();
        else:
          iprot.skip(ftype)
      elif fid == 7:
        if ftype == TType.STRUCT:
          self.last_error_data = RecordErrorRequest()
          self.last_error_data.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('ErrorInfo')
    if self.error_count is not None:
      oprot.writeFieldBegin('error_count', TType.I64, 1)
      oprot.writeI64(self.error_count)
      oprot.writeFieldEnd()
    if self.developer_email is not None:
      oprot.writeFieldBegin('developer_email', TType.STRING, 2)
      oprot.writeString(self.developer_email)
      oprot.writeFieldEnd()
    if self.date is not None:
      oprot.writeFieldBegin('date', TType.STRING, 3)
      oprot.writeString(self.date)
      oprot.writeFieldEnd()
    if self.email_sent is not None:
      oprot.writeFieldBegin('email_sent', TType.BOOL, 4)
      oprot.writeBool(self.email_sent)
      oprot.writeFieldEnd()
    if self.last_occurrence is not None:
      oprot.writeFieldBegin('last_occurrence', TType.STRING, 5)
      oprot.writeString(self.last_occurrence)
      oprot.writeFieldEnd()
    if self.is_known_error is not None:
      oprot.writeFieldBegin('is_known_error', TType.BOOL, 6)
      oprot.writeBool(self.is_known_error)
      oprot.writeFieldEnd()
    if self.last_error_data is not None:
      oprot.writeFieldBegin('last_error_data', TType.STRUCT, 7)
      self.last_error_data.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class EmailRemapping(object):
  """
  Attributes:
   - remap
   - last_update_ts
  """

  thrift_spec = (
    None, # 0
    (1, TType.MAP, 'remap', (TType.STRING,None,TType.STRING,None), {
    }, ), # 1
    (2, TType.I64, 'last_update_ts', None, None, ), # 2
  )

  def __init__(self, remap=thrift_spec[1][4], last_update_ts=None,):
    if remap is self.thrift_spec[1][4]:
      remap = {
    }
    self.remap = remap
    self.last_update_ts = last_update_ts

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.MAP:
          self.remap = {}
          (_ktype17, _vtype18, _size16 ) = iprot.readMapBegin() 
          for _i20 in xrange(_size16):
            _key21 = iprot.readString();
            _val22 = iprot.readString();
            self.remap[_key21] = _val22
          iprot.readMapEnd()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I64:
          self.last_update_ts = iprot.readI64();
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('EmailRemapping')
    if self.remap is not None:
      oprot.writeFieldBegin('remap', TType.MAP, 1)
      oprot.writeMapBegin(TType.STRING, TType.STRING, len(self.remap))
      for kiter23,viter24 in self.remap.items():
        oprot.writeString(kiter23)
        oprot.writeString(viter24)
      oprot.writeMapEnd()
      oprot.writeFieldEnd()
    if self.last_update_ts is not None:
      oprot.writeFieldBegin('last_update_ts', TType.I64, 2)
      oprot.writeI64(self.last_update_ts)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class WatchFileEntry(object):
  """
  Attributes:
   - email
   - filepath
   - watch_all_errors
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'email', None, None, ), # 1
    (2, TType.STRING, 'filepath', None, None, ), # 2
    (3, TType.BOOL, 'watch_all_errors', None, None, ), # 3
  )

  def __init__(self, email=None, filepath=None, watch_all_errors=None,):
    self.email = email
    self.filepath = filepath
    self.watch_all_errors = watch_all_errors

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.email = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.filepath = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.BOOL:
          self.watch_all_errors = iprot.readBool();
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('WatchFileEntry')
    if self.email is not None:
      oprot.writeFieldBegin('email', TType.STRING, 1)
      oprot.writeString(self.email)
      oprot.writeFieldEnd()
    if self.filepath is not None:
      oprot.writeFieldBegin('filepath', TType.STRING, 2)
      oprot.writeString(self.filepath)
      oprot.writeFieldEnd()
    if self.watch_all_errors is not None:
      oprot.writeFieldBegin('watch_all_errors', TType.BOOL, 3)
      oprot.writeBool(self.watch_all_errors)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class WatchList(object):
  """
  Attributes:
   - watches
   - last_update_ts
  """

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'watches', (TType.STRUCT,(WatchFileEntry, WatchFileEntry.thrift_spec)), [
    ], ), # 1
    (2, TType.I64, 'last_update_ts', None, None, ), # 2
  )

  def __init__(self, watches=thrift_spec[1][4], last_update_ts=None,):
    if watches is self.thrift_spec[1][4]:
      watches = [
    ]
    self.watches = watches
    self.last_update_ts = last_update_ts

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.LIST:
          self.watches = []
          (_etype28, _size25) = iprot.readListBegin()
          for _i29 in xrange(_size25):
            _elem30 = WatchFileEntry()
            _elem30.read(iprot)
            self.watches.append(_elem30)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I64:
          self.last_update_ts = iprot.readI64();
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('WatchList')
    if self.watches is not None:
      oprot.writeFieldBegin('watches', TType.LIST, 1)
      oprot.writeListBegin(TType.STRUCT, len(self.watches))
      for iter31 in self.watches:
        iter31.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.last_update_ts is not None:
      oprot.writeFieldBegin('last_update_ts', TType.I64, 2)
      oprot.writeI64(self.last_update_ts)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class KnownError(object):
  """
  Attributes:
   - filename
   - function_name
   - code_fragment
   - min_alert_threshold
   - max_alert_threshold
   - email_recipients
   - email_header
   - alert_every_n_occurences
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'filename', None, None, ), # 1
    (2, TType.STRING, 'function_name', None, None, ), # 2
    (3, TType.STRING, 'code_fragment', None, None, ), # 3
    (4, TType.I64, 'min_alert_threshold', None, None, ), # 4
    (5, TType.I64, 'max_alert_threshold', None, None, ), # 5
    (6, TType.LIST, 'email_recipients', (TType.STRING,None), None, ), # 6
    (7, TType.STRING, 'email_header', None, None, ), # 7
    (8, TType.I64, 'alert_every_n_occurences', None, None, ), # 8
  )

  def __init__(self, filename=None, function_name=None, code_fragment=None, min_alert_threshold=None, max_alert_threshold=None, email_recipients=None, email_header=None, alert_every_n_occurences=None,):
    self.filename = filename
    self.function_name = function_name
    self.code_fragment = code_fragment
    self.min_alert_threshold = min_alert_threshold
    self.max_alert_threshold = max_alert_threshold
    self.email_recipients = email_recipients
    self.email_header = email_header
    self.alert_every_n_occurences = alert_every_n_occurences

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.filename = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.function_name = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.STRING:
          self.code_fragment = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.I64:
          self.min_alert_threshold = iprot.readI64();
        else:
          iprot.skip(ftype)
      elif fid == 5:
        if ftype == TType.I64:
          self.max_alert_threshold = iprot.readI64();
        else:
          iprot.skip(ftype)
      elif fid == 6:
        if ftype == TType.LIST:
          self.email_recipients = []
          (_etype35, _size32) = iprot.readListBegin()
          for _i36 in xrange(_size32):
            _elem37 = iprot.readString();
            self.email_recipients.append(_elem37)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 7:
        if ftype == TType.STRING:
          self.email_header = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 8:
        if ftype == TType.I64:
          self.alert_every_n_occurences = iprot.readI64();
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('KnownError')
    if self.filename is not None:
      oprot.writeFieldBegin('filename', TType.STRING, 1)
      oprot.writeString(self.filename)
      oprot.writeFieldEnd()
    if self.function_name is not None:
      oprot.writeFieldBegin('function_name', TType.STRING, 2)
      oprot.writeString(self.function_name)
      oprot.writeFieldEnd()
    if self.code_fragment is not None:
      oprot.writeFieldBegin('code_fragment', TType.STRING, 3)
      oprot.writeString(self.code_fragment)
      oprot.writeFieldEnd()
    if self.min_alert_threshold is not None:
      oprot.writeFieldBegin('min_alert_threshold', TType.I64, 4)
      oprot.writeI64(self.min_alert_threshold)
      oprot.writeFieldEnd()
    if self.max_alert_threshold is not None:
      oprot.writeFieldBegin('max_alert_threshold', TType.I64, 5)
      oprot.writeI64(self.max_alert_threshold)
      oprot.writeFieldEnd()
    if self.email_recipients is not None:
      oprot.writeFieldBegin('email_recipients', TType.LIST, 6)
      oprot.writeListBegin(TType.STRING, len(self.email_recipients))
      for iter38 in self.email_recipients:
        oprot.writeString(iter38)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.email_header is not None:
      oprot.writeFieldBegin('email_header', TType.STRING, 7)
      oprot.writeString(self.email_header)
      oprot.writeFieldEnd()
    if self.alert_every_n_occurences is not None:
      oprot.writeFieldBegin('alert_every_n_occurences', TType.I64, 8)
      oprot.writeI64(self.alert_every_n_occurences)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class KnownErrorList(object):
  """
  Attributes:
   - identifiers
   - last_update_ts
  """

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'identifiers', (TType.STRUCT,(KnownError, KnownError.thrift_spec)), [
    ], ), # 1
    (2, TType.I64, 'last_update_ts', None, None, ), # 2
  )

  def __init__(self, identifiers=thrift_spec[1][4], last_update_ts=None,):
    if identifiers is self.thrift_spec[1][4]:
      identifiers = [
    ]
    self.identifiers = identifiers
    self.last_update_ts = last_update_ts

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.LIST:
          self.identifiers = []
          (_etype42, _size39) = iprot.readListBegin()
          for _i43 in xrange(_size39):
            _elem44 = KnownError()
            _elem44.read(iprot)
            self.identifiers.append(_elem44)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I64:
          self.last_update_ts = iprot.readI64();
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('KnownErrorList')
    if self.identifiers is not None:
      oprot.writeFieldBegin('identifiers', TType.LIST, 1)
      oprot.writeListBegin(TType.STRUCT, len(self.identifiers))
      for iter45 in self.identifiers:
        iter45.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.last_update_ts is not None:
      oprot.writeFieldBegin('last_update_ts', TType.I64, 2)
      oprot.writeI64(self.last_update_ts)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class CodeIdentifier(object):
  """
  Attributes:
   - filename
   - function_name
   - code_fragment
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'filename', None, None, ), # 1
    (2, TType.STRING, 'function_name', None, None, ), # 2
    (3, TType.STRING, 'code_fragment', None, None, ), # 3
  )

  def __init__(self, filename=None, function_name=None, code_fragment=None,):
    self.filename = filename
    self.function_name = function_name
    self.code_fragment = code_fragment

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.filename = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.function_name = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.STRING:
          self.code_fragment = iprot.readString();
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('CodeIdentifier')
    if self.filename is not None:
      oprot.writeFieldBegin('filename', TType.STRING, 1)
      oprot.writeString(self.filename)
      oprot.writeFieldEnd()
    if self.function_name is not None:
      oprot.writeFieldBegin('function_name', TType.STRING, 2)
      oprot.writeString(self.function_name)
      oprot.writeFieldEnd()
    if self.code_fragment is not None:
      oprot.writeFieldBegin('code_fragment', TType.STRING, 3)
      oprot.writeString(self.code_fragment)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class CodeIdentifierList(object):
  """
  Attributes:
   - identifiers
   - last_update_ts
  """

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'identifiers', (TType.STRUCT,(CodeIdentifier, CodeIdentifier.thrift_spec)), [
    ], ), # 1
    (2, TType.I64, 'last_update_ts', None, None, ), # 2
  )

  def __init__(self, identifiers=thrift_spec[1][4], last_update_ts=None,):
    if identifiers is self.thrift_spec[1][4]:
      identifiers = [
    ]
    self.identifiers = identifiers
    self.last_update_ts = last_update_ts

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.LIST:
          self.identifiers = []
          (_etype49, _size46) = iprot.readListBegin()
          for _i50 in xrange(_size46):
            _elem51 = CodeIdentifier()
            _elem51.read(iprot)
            self.identifiers.append(_elem51)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I64:
          self.last_update_ts = iprot.readI64();
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('CodeIdentifierList')
    if self.identifiers is not None:
      oprot.writeFieldBegin('identifiers', TType.LIST, 1)
      oprot.writeListBegin(TType.STRUCT, len(self.identifiers))
      for iter52 in self.identifiers:
        iter52.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.last_update_ts is not None:
      oprot.writeFieldBegin('last_update_ts', TType.I64, 2)
      oprot.writeI64(self.last_update_ts)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

import copy
import datetime
import email
import time
import unittest
import StringIO

from flawless.lib.storage.stub import StubStorage
import flawless.lib.config
import flawless.server.api.ttypes as api_ttypes
from flawless.server.service import FlawlessThriftServiceHandler
import flawless.server.server


class BaseTestCase(unittest.TestCase):

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self.popen_stub = POpenStub()
        self.popen_stub.stdout = StringIO.StringIO(
            "75563df6e9d1efe44b48f6643fde9ebbd822b0c5 25 25 1\n"
            "author John Egan\n"
            "author-mail <wishbone@shopkick.com>\n"
            "author-time %d\n"
            "author-tz -0800\n"
            "committer John Egan\n"
            "committer-mail <repo_master@shopkick.com>\n"
            "committer-time 1356245776\n"
            "committer-tz -0800\n"
            "summary Add more robust support for string keys\n"
            "previous 3491c7b8e298ec81dc7583163a118e7c2250999f safe_access.py\n"
            "filename safe_access.py\n"
            "                             ex: a.b.[myvar] where myvar is passed in as a kwarg\n"
            % int(time.mktime(datetime.datetime(2017, 7, 30).timetuple()))
        )

        self._set_stub_time(datetime.datetime(2020, 1, 1))
        self.smtp_stub = SMTPClientStub()
        self.file_open_stub = OpenFileStub()

        self.watchers = api_ttypes.WatchList([
            api_ttypes.WatchFileEntry(
                email="wilfred@shopkick.com",
                filepath="tools/furminator.py",
                watch_all_errors=True,
            ),

            api_ttypes.WatchFileEntry(
                email="lassie@shopkick.com",
                filepath="tools/furminator.py",
                watch_all_errors=True,
            ),
            api_ttypes.WatchFileEntry(
                email="wishbone@shopkick.com",
                filepath="lib/no/such/path/for/testing.py",
                watch_all_errors=True,
            ),
        ])

        self.third_party_whitelist = api_ttypes.CodeIdentifierList([
            api_ttypes.CodeIdentifier('facebook.py', 'post_treat_to_facebook',
                                      'urllib.urlencode(args), post_data)'),
            api_ttypes.CodeIdentifier(
                'SQLAlchemy-0.5.6-py2.6.egg/sqlalchemy/pool.py',
                'do_get',
                'raise exc.TimeoutError("QueuePool limit of size %d overflow %d ',
            ),
        ])

        self.known_errors = api_ttypes.KnownErrorList([
            api_ttypes.KnownError(
                'lib/doghouse/authentication.py',
                'check_auth',
                'raise errors.BadAuthenticationError("Something smells off...")',
                max_alert_threshold=0,
            ),
            api_ttypes.KnownError(
                filename='coreservices/waterbowl/rewards/water.py',
                function_name='make_external_api_ttypes_request',
                code_fragment='raise api_ttypes.WaterbowlError(error_code='
                              'api_ttypes.WaterbowlErrorCode.OUT_OF_WATER, message=str(e))',
                email_recipients=["wilfred@shopkick.com", "snoopy@shopkick.com"],
                email_header='NOTE: This error typically does not require dev team involvement.',
                alert_every_n_occurences=1,
            ),
        ])

        self.building_blocks = api_ttypes.CodeIdentifierList([
            api_ttypes.CodeIdentifier('apps/shopkick/doghouse/lib/base.py',
                                      '_get_request_param',
                                      'raise errors.BadRequestError("Missing param %s" % name)'),
        ])

        self.config_storage_stub = StubStorage(partition=None)
        self.config_storage_stub["watch_list"] = self.watchers
        self.config_storage_stub["third_party_whitelist"] = self.third_party_whitelist
        self.config_storage_stub["known_errors"] = self.known_errors
        self.config_storage_stub["building_blocks"] = self.building_blocks
        self.errors_storage_stub = StubStorage(partition=None)

        self.saved_config = copy.deepcopy(flawless.lib.config.get().__dict__)
        self.test_config = flawless.lib.config.get()
        self.test_config.__dict__ = dict((o.name, o.default) for o in flawless.lib.config.OPTIONS)
        self.test_config.repo_dir = "/tmp"
        self.test_config.report_only_after_minimum_date = "2010-01-01"
        self.test_config.report_error_threshold = 1
        self.test_config.only_blame_filepaths_matching = [
            r"^coreservices(?!.*/thrift/).*$",
            r"lib/.*",
            r"tools/.*",
        ]
        self.test_config.report_runtime_package_directory_names = ["site-packages"]
        self.test_config.config_dir_path = "../config"

        self.handler = FlawlessThriftServiceHandler(
            open_process_func=self.popen_stub,
            storage_factory=lambda partition: self.errors_storage_stub if partition else self.config_storage_stub,
            smtp_client_cls=self.smtp_stub,
            time_func=lambda: self.stub_time,
            thread_cls=ThreadStub,
        )

    def tearDown(self):
        super(BaseTestCase, self).tearDown()
        flawless.lib.config.get().__dict__ = self.saved_config

    def _set_stub_time(self, dt):
        self.stub_time = int(time.mktime(dt.timetuple()))

    def assertDictEquals(self, expected, actual):
        if expected == actual:
            return

        bad_keys = set(expected.keys()) ^ set(actual.keys())
        if bad_keys:
            keys_str = [str(key) for key in (set(expected.keys()) - set(actual.keys()))]
            errstr = "Keys not in second dict: %s" % "\n".join(keys_str)
            keys_str = [str(key) for key in (set(actual.keys()) - set(expected.keys()))]
            errstr += "\nKeys not in first dict:    %s" % "\n".join(keys_str)
            raise AssertionError("Missing/Extraneous keys:\n%s" % errstr)
        for key in expected.keys():
            if expected[key] != actual[key]:
                raise AssertionError("Value mismatch for key %s:\n%s\n%s" %
                                     (key, str(expected[key]), str(actual[key])))

    def assertEmailEquals(self, expected, actual):
        self.assertEquals(expected["from_address"], actual["from_address"])
        self.assertEquals(set(expected["to_addresses"]), set(actual["to_addresses"]))
        parsed_email = email.message_from_string(actual["body"])
        self.assertEquals(expected["from_address"], parsed_email["From"])
        self.assertEquals(set(expected["to_addresses"]), set(parsed_email["To"].split(", ")))
        self.assertEquals(expected["subject"], parsed_email["Subject"])
        self.assertTrue(expected["body"] in parsed_email.get_payload(decode=True))


class RecordErrorTestCase(BaseTestCase):

    def setUp(self):
        super(RecordErrorTestCase, self).setUp()

    def test_records_error(self):
        req = api_ttypes.RecordErrorRequest(
            traceback=[api_ttypes.StackLine("/site-packages/lib/test.py", 5, "test_func", "code"),
                       api_ttypes.StackLine("/site-packages/coreservices/service.py", 7, "serve", "..."),
                       api_ttypes.StackLine("/site-packages/thirdparty/3rdparty_lib.py", 9, "call", "x")],
            exception_message="email text",
            hostname="localhost",
        )

        self.handler.record_error(req)
        self.assertDictEquals({
            api_ttypes.ErrorKey("coreservices/service.py", 7, "serve", "..."): api_ttypes.ErrorInfo(
                1, "wishbone@shopkick.com", "2017-07-30 00:00:00", True, "2020-01-01 00:00:00",
                is_known_error=False, last_error_data=req)},
            self.handler.errors_seen.dict)
        self.assertEqual(["git", "--git-dir=/tmp/.git", "--work-tree=/tmp", "blame",
                          "-p", "/tmp/coreservices/service.py", "-L", "7,+1"],
                         self.popen_stub.last_args)
        self.assertEmailEquals(dict(to_addresses=["wishbone@shopkick.com"],
                                    from_address="flawless@example.com",
                                    subject="Error on localhost in coreservices/service.py",
                                    body="email text",
                                    smtp_server_host_port=None),
                               self.smtp_stub.last_args)

    def test_records_error_with_thrift_in_file_name(self):
        req = api_ttypes.RecordErrorRequest(
            traceback=[api_ttypes.StackLine("/site-packages/coreservices/thrift_file.py", 7, "serve", "..."),
                       api_ttypes.StackLine("/site-packages/thirdparty/3rdparty_lib.py", 9, "call", "x")],
            exception_message="email text",
            hostname="localhost",
        )

        self.handler.record_error(req)
        self.assertDictEquals({
            api_ttypes.ErrorKey("coreservices/thrift_file.py", 7, "serve", "..."): api_ttypes.ErrorInfo(
                1, "wishbone@shopkick.com", "2017-07-30 00:00:00", True, "2020-01-01 00:00:00",
                is_known_error=False, last_error_data=req)},
            self.handler.errors_seen.dict)

    def test_ignores_error_in_thrift_directory(self):
        req = api_ttypes.RecordErrorRequest(
            traceback=[api_ttypes.StackLine("/site-packages/thirdparty/3rdparty_lib.py", 9, "call", "x"),
                       api_ttypes.StackLine("/site-packages/coreservices/thrift/file.py", 7, "serve", "...")],
            exception_message="email text",
            hostname="localhost",
        )

        self.handler.record_error(req)
        self.assertDictEquals({}, self.handler.errors_seen.dict)

    def test_doesnt_report_errors_under_threshold(self):
        self.test_config.report_error_threshold = 2
        req = api_ttypes.RecordErrorRequest(
            traceback=[api_ttypes.StackLine("/site-packages/lib/test.py", 5, "test_func", "code"),
                       api_ttypes.StackLine("/site-packages/coreservices/service.py", 7, "serve", "..."),
                       api_ttypes.StackLine("/site-packages/thirdparty/3rdparty_lib.py", 9, "call", "x")],
            exception_message="email text",
            hostname="localhost",
        )

        self.handler.record_error(req)
        self.assertDictEquals({
            api_ttypes.ErrorKey("coreservices/service.py", 7, "serve", "..."): api_ttypes.ErrorInfo(
                1, "wishbone@shopkick.com", "2017-07-30 00:00:00", False, "2020-01-01 00:00:00",
                is_known_error=False, last_error_data=req)},
            self.handler.errors_seen.dict)
        self.assertEquals(None, self.smtp_stub.last_args)

    def test_uses_threshold_specified_in_request(self):
        self.test_config.report_error_threshold = 2
        req = api_ttypes.RecordErrorRequest(
            traceback=[api_ttypes.StackLine("/site-packages/lib/test.py", 5, "test_func", "code"),
                       api_ttypes.StackLine("/site-packages/coreservices/service.py", 7, "serve", "..."),
                       api_ttypes.StackLine("/site-packages/thirdparty/3rdparty_lib.py", 9, "call", "x")],
            exception_message="email text",
            hostname="localhost",
            error_threshold=1,
        )

        self.handler.record_error(req)

        self.assertEmailEquals(dict(to_addresses=["wishbone@shopkick.com"],
                                    from_address="flawless@example.com",
                                    cc_address=None,
                                    bcc_address=None,
                                    subject="Error on localhost in coreservices/service.py",
                                    body="email text",
                                    smtp_server_host_port=None),
                               self.smtp_stub.last_args)

    def test_always_alerts_on_red_alert_errors(self):
        self.test_config.report_error_threshold = 3
        req = api_ttypes.RecordErrorRequest(
            traceback=[api_ttypes.StackLine("/site-packages/lib/test.py", 5, "test_func", "code"),
                       api_ttypes.StackLine("/site-packages/coreservices/service.py", 7, "serve", "..."),
                       api_ttypes.StackLine('/site-packages/coreservices/service/utils.py',
                                            9, "check_water_levels", '% (waterbowl_id, min_required_level))'),
                       api_ttypes.StackLine("/site-packages/coreservices/waterbowl/rewards/water.py",
                                            5,
                                            "make_external_api_ttypes_request",
                                            "raise api_ttypes.WaterbowlError(error_code=api_ttypes.WaterbowlErrorCode."
                                            "OUT_OF_WATER, message=str(e))")],
            exception_message="email text",
            hostname="localhost",
        )

        self.handler.record_error(req)

        # The 2 red alert recipients plus the developer responsible
        self.assertEmailEquals(dict(to_addresses=["wilfred@shopkick.com", "wishbone@shopkick.com",
                                                  "snoopy@shopkick.com"],
                                    from_address="flawless@example.com",
                                    cc_address=None,
                                    bcc_address=None,
                                    subject="Error on localhost in coreservices/waterbowl/rewards/water.py",
                                    body="email text",
                                    smtp_server_host_port=None),
                               self.smtp_stub.last_args)

    def test_email_includes_watchers(self):
        # Almost the same setup as test_uses_threshold_specified_in_request except the traceback includes
        # a path that Yen is watching
        self.test_config.report_error_threshold = 2
        req = api_ttypes.RecordErrorRequest(
            traceback=[api_ttypes.StackLine("/site-packages/lib/test.py", 5, "test_func", "code"),
                       api_ttypes.StackLine("/site-packages/tools/furminator.py", 7, "fubar", "..."),
                       api_ttypes.StackLine("/site-packages/thirdparty/3rdparty_lib.py", 9, "call", "x")],
            exception_message="email text",
            hostname="localhost",
            error_threshold=1,
        )

        self.handler.record_error(req)
        self.assertEmailEquals(dict(to_addresses=["wilfred@shopkick.com", "wishbone@shopkick.com", "lassie@shopkick.com"],
                                    from_address="flawless@example.com",
                                    cc_address=None,
                                    bcc_address=None,
                                    subject="Error on localhost in tools/furminator.py",
                                    body="email text",
                                    smtp_server_host_port=None),
                               self.smtp_stub.last_args)

    def test_email_includes_extra_information(self):
        # Traceback includes a path that has extra_information tagged on it
        self.test_config.report_error_threshold = 2
        req = api_ttypes.RecordErrorRequest(
            traceback=[
                api_ttypes.StackLine(
                    "/site-packages/coreservices/waterbowl/rewards/water.py",
                    5,
                    "make_external_api_ttypes_request",
                    "raise api_ttypes.WaterbowlError(error_code=api_ttypes.WaterbowlErrorCode."
                    "OUT_OF_WATER, message=str(e))"
                ),
            ],
            exception_message="email text",
            hostname="localhost",
            error_threshold=1,
            additional_info="extra stuff",
        )

        self.handler.record_error(req)

        self.assertEmailEquals(dict(to_addresses=["snoopy@shopkick.com",
                                                  "wilfred@shopkick.com",
                                                  "wishbone@shopkick.com"],
                                    from_address="flawless@example.com",
                                    cc_address=None,
                                    bcc_address=None,
                                    subject="Error on localhost in coreservices/waterbowl/rewards/water.py",
                                    body='NOTE: This error typically does not require dev team involvement.',
                                    smtp_server_host_port=None),
                               self.smtp_stub.last_args)
        body = email.message_from_string(self.smtp_stub.last_args["body"]).get_payload(decode=True)
        self.assertTrue("email text" in body)
        self.assertTrue("extra stuff" in body)

    def test_removes_duplicate_emails(self):
        # Almost the same setup as test_uses_threshold_specified_in_request except the traceback includes
        # a path that John is watching. Since he's also the developer who committed the buggy code, he is
        # the one and only email recipient.
        self.test_config.report_error_threshold = 2
        req = api_ttypes.RecordErrorRequest(
            traceback=[api_ttypes.StackLine("/site-packages/lib/test.py", 5, "test_func", "code"),
                       api_ttypes.StackLine("/site-packages/lib/no/such/path/for/testing.py", 7, "fubar", "..."),
                       api_ttypes.StackLine("/site-packages/thirdparty/3rdparty_lib.py", 9, "call", "x")],
            exception_message="email text",
            hostname="localhost",
            error_threshold=1,
        )

        self.handler.record_error(req)

        self.assertEmailEquals(dict(to_addresses=["wishbone@shopkick.com"],
                                    from_address="flawless@example.com",
                                    cc_address=None,
                                    bcc_address=None,
                                    subject="Error on localhost in lib/no/such/path/for/testing.py",
                                    body="email text",
                                    smtp_server_host_port=None),
                               self.smtp_stub.last_args)

    def test_doesnt_email_on_errors_before_cutoff_date(self):
        req = api_ttypes.RecordErrorRequest(
            traceback=[api_ttypes.StackLine("/site-packages/lib/test.py", 5, "test_func", "code"),
                       api_ttypes.StackLine("/site-packages/coreservices/service.py", 7, "serve", "..."),
                       api_ttypes.StackLine("/site-packages/thirdparty/3rdparty_lib.py", 9, "call", "x")],
            exception_message="email text",
            hostname="localhost",
        )
        self.popen_stub.stdout = StringIO.StringIO(
            "75563df6e9d1efe44b48f6643fde9ebbd822b0c5 25 25 1\n"
            "author John Egan\n"
            "author-mail <wishbone@shopkick.com>\n"
            "author-time %d\n"
            "author-tz -0800\n"
            % int(time.mktime(datetime.datetime(2009, 7, 30).timetuple()))
        )

        self.handler.record_error(req)
        self.assertDictEquals({
            api_ttypes.ErrorKey("coreservices/service.py", 7, "serve", "..."): api_ttypes.ErrorInfo(
                1, "wishbone@shopkick.com", "2009-07-30 00:00:00", False,
                "2020-01-01 00:00:00", is_known_error=False, last_error_data=req)},
            self.handler.errors_seen.dict)
        self.assertEqual(None, self.smtp_stub.last_args)

    def test_records_error_only_once(self):
        req = api_ttypes.RecordErrorRequest(
            traceback=[api_ttypes.StackLine("/site-packages/lib/test.py", 5, "test_func", "code"),
                       api_ttypes.StackLine("/site-packages/coreservices/service.py", 7, "serve", "..."),
                       api_ttypes.StackLine("/site-packages/thirdparty/3rdparty_lib.py", 9, "call", "x")],
            exception_message="email text",
            hostname="localhost",
        )
        self.handler.record_error(req)
        self._set_stub_time(datetime.datetime(2020, 1, 2))
        self.handler.record_error(req)

        self.assertDictEquals({
            api_ttypes.ErrorKey("coreservices/service.py", 7, "serve", "..."): api_ttypes.ErrorInfo(
                2, "wishbone@shopkick.com", "2017-07-30 00:00:00", True, "2020-01-02 00:00:00",
                is_known_error=False, last_error_data=req)},
            self.handler.errors_seen.dict)
        self.assertEqual(1, len(self.smtp_stub.args_list))

    def test_does_not_email_for_whitelisted_errors(self):
        req = api_ttypes.RecordErrorRequest(
            traceback=[api_ttypes.StackLine("/site-packages/lib/test.py", 5, "test_func", "code"),
                       api_ttypes.StackLine("/site-packages/lib/doghouse/authentication.py", 7, "check_auth",
                                            'raise errors.BadAuthenticationError("Something smells off...")')],
            exception_message="email text",
            hostname="localhost",
        )
        self.handler.record_error(req)

        self.assertDictEquals({
            api_ttypes.ErrorKey("lib/doghouse/authentication.py", 7, "check_auth",
                                'raise errors.BadAuthenticationError("Something smells off...")'
                                ): api_ttypes.ErrorInfo(1, "wishbone@shopkick.com", "2017-07-30 00:00:00",
                                                        False, "2020-01-01 00:00:00", is_known_error=True,
                                                        last_error_data=req)},
                              self.handler.errors_seen.dict)
        self.assertEquals(None, self.smtp_stub.last_args)

    def test_ignores_third_party_whitelisted_errors(self):
        req = api_ttypes.RecordErrorRequest(
            traceback=[api_ttypes.StackLine("/site-packages/lib/test.py", 5, "test_func", "code"),
                       api_ttypes.StackLine("/site-packages/SQLAlchemy-0.5.6-py2.6.egg/sqlalchemy/pool.py",
                                            7,
                                            "do_get",
                                            'raise exc.TimeoutError("QueuePool limit of size %d overflow %d '
                                            'reached, connection timed out, timeout %d" % (self.size(),'
                                            'self.overflow(), self._timeout))')],
            exception_message="email text",
            hostname="localhost",
        )
        self.handler.record_error(req)
        self.assertDictEquals({}, self.handler.errors_seen.dict)

    def test_ignores_third_party_whitelisted_errors_for_facebook(self):
        req = api_ttypes.RecordErrorRequest(
            traceback=[api_ttypes.StackLine("/site-packages/lib/test.py", 5, "test_func", "code"),
                       api_ttypes.StackLine("/site-packages/facebook.py",
                                            7,
                                            "post_treat_to_facebook",
                                            'urllib.urlencode(args), post_data)')],
            exception_message="email text",
            hostname="localhost",
        )
        self.handler.record_error(req)
        self.assertDictEquals({}, self.handler.errors_seen.dict)

    def test_traces_up_stack_trace_for_errors_originating_from_building_blocks(self):
        req = api_ttypes.RecordErrorRequest(
            traceback=[api_ttypes.StackLine("/site-packages/lib/test.py", 5, "test_func", "code"),
                       api_ttypes.StackLine("/site-packages/coreservices/service.py", 7, "serve", "..."),
                       api_ttypes.StackLine("/site-packages/apps/shopkick/doghouse/lib/base.py", 9,
                                            "_get_request_param",
                                            'raise errors.BadRequestError("Missing param %s" % name)')],
            exception_message="email text",
            hostname="localhost",
        )
        self.handler.record_error(req)

        self.assertDictEquals({
            api_ttypes.ErrorKey("coreservices/service.py", 7, "serve", "..."): api_ttypes.ErrorInfo(
                1, "wishbone@shopkick.com", "2017-07-30 00:00:00", True, "2020-01-01 00:00:00",
                is_known_error=False, last_error_data=req)},
            self.handler.errors_seen.dict)


############################## Stubs ##############################
class LogStub(object):

    def __init__(self):
        self.last_args = None
        self.args_list = []

    def info(self, args):
        self.last_args = args
        self.args_list.append(args)


class POpenStub(object):

    def __init__(self):
        self.last_args = None
        self.stdout = StringIO.StringIO()
        self.stderr = StringIO.StringIO()

    def __call__(self, args, **kwargs):
        self.last_args = args
        return self


class SMTPClientStub(object):
    def __init__(self):
        self.args_list = []
        self.last_args = None
        self.host = None
        self.port = None

    def __call__(self, host, port):
        self.host = host
        self.port = port
        return self

    def sendmail(self, from_address, to_addresses, body):
        self.last_args = dict((k, v) for k, v in locals().items() if k != 'self')
        self.args_list.append(self.last_args)

    def quit(self):
        pass

    def login(self, user, password):
        pass


class OpenFileStub(object):
    def __init__(self):
        self.files = dict()
        self.current_file = None

    def set_file(self, filename, contents):
        self.files[filename] = StringIO.StringIO(contents)

    def __enter__(self, *args, **kwargs):
        return self.files[self.current_file]

    def __exit__(self, type, value, traceback):
        pass

    def __call__(self, filename, *args, **kwargs):
        self.current_file = filename
        return self


class ThreadStub(object):
    def __init__(self, target=None, args=[]):
        self.target = target
        self.args = args

    def start(self):
        if self.target.__name__ != '_run_background_update_thread':
            self.target(*self.args)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

import copy
import time
import unittest


import flawless.client
import flawless.client.client
import flawless.client.decorators
from flawless.server.stub import FlawlessServiceStub


@flawless.client.decorators.wrap_class
class ThriftTestHandler(object):

    def __init__(self):
        self.instancevar = 98

    def method(self, fail=False, result=42):
        return self._simulate_call(fail, result)

    @classmethod
    def classmeth(cls, fail=False, result=43):
        return cls._simulate_call(fail, result)

    @classmethod
    def _simulate_call(cls, fail=False, result=42):
        if fail:
            raise Exception()

        return result

    def check_health(self, fail=False, result=42, delay=0):
        return self._simulate_call(fail=fail, result=result, delay=delay)


class BaseErrorsTestCase(unittest.TestCase):

    def setUp(self):
        self.client_stub = FlawlessServiceStub()
        self.saved_get_get_service = flawless.client.client._get_service
        setattr(flawless.client.client, "_get_service", lambda: (self.client_stub, TransportStub()))
        self.saved_config = copy.deepcopy(flawless.lib.config.get().__dict__)
        self.test_config = flawless.lib.config.get()
        self.test_config.__dict__ = dict((o.name, o.default) for o in flawless.lib.config.OPTIONS)
        flawless.client.set_hostports(["localhost:9028"])

    def tearDown(self):
        setattr(flawless.client.client, "_get_service", self.saved_get_get_service)
        flawless.lib.config.get().__dict__ = self.saved_config


class ClassDecoratorTestCase(BaseErrorsTestCase):
    def setUp(self):
        super(ClassDecoratorTestCase, self).setUp()
        self.handler = ThriftTestHandler()

    def test_returns_correct_result(self):
        self.assertEquals(56, self.handler.method(result=56))

    def test_returns_correct_result_for_classmethod(self):
        self.assertEquals(91, ThriftTestHandler.classmeth(result=91))

    def test_should_call_flawless_backend_on_exception(self):
        self.assertRaises(Exception, self.handler.method, fail=True)
        self.assertEquals(1, len(self.client_stub.record_error.args_list))
        errorFound = False
        req_obj = self.client_stub.record_error.last_args['request']
        for row in req_obj.traceback:
            if row.function_name == "_simulate_call":
                errorFound = True
            if row.function_name == "method":
                self.assertEquals('98', row.frame_locals['self.instancevar'])
        self.assertTrue(errorFound)
        self.assertEqual(None, req_obj.error_threshold)

    def test_logs_classvars(self):
        self.assertRaises(Exception, self.handler.method, fail=True)
        self.assertEquals(1, len(self.client_stub.record_error.args_list))
        errorFound = False
        req_obj = self.client_stub.record_error.last_args['request']
        for row in req_obj.traceback:
            if row.function_name == "_simulate_call":
                errorFound = True
        self.assertTrue(errorFound)
        self.assertEqual(None, req_obj.error_threshold)


class FunctionDecoratorTestCase(BaseErrorsTestCase):
    def setUp(self):
        super(FunctionDecoratorTestCase, self).setUp()

    @flawless.client.decorators.wrap_function
    def example_func(self, fail=False, retval=None):
        myvar = 7
        if fail:
            raise Exception(":(")
        return retval

    @flawless.client.decorators.wrap_function(error_threshold=7, reraise_exception=False)
    def second_example_func(self, fail=False, retval=None):
        if fail:
            raise Exception("woohoo")
        return retval

    def test_returns_correct_result(self):
        self.assertEquals(7, self.example_func(fail=False, retval=7))

    def test_should_call_flawless_backend_on_exception(self):
        self.assertRaises(Exception, self.example_func, fail=True)
        errorFound = False
        req_obj = self.client_stub.record_error.last_args['request']
        for row in req_obj.traceback:
            if row.function_name == "example_func":
                errorFound = True
        self.assertTrue(errorFound)
        self.assertEqual(None, req_obj.error_threshold)

    def test_does_not_call_flawless_if_backoff(self):
        flawless.client.client.BACKOFF_MS = int(time.time() * 1000) + 1000
        self.assertRaises(Exception, self.example_func, fail=True)
        self.assertEqual(None, self.client_stub.record_error.last_args)
        flawless.client.client.BACKOFF_MS = 0

    def test_decorator_with_kwargs(self):
        self.second_example_func(fail=True)
        errorFound = False
        req_obj = self.client_stub.record_error.last_args['request']
        for row in req_obj.traceback:
            if row.function_name == "second_example_func":
                errorFound = True
        self.assertTrue(errorFound)
        self.assertEqual(7, req_obj.error_threshold)

    def test_logs_locals(self):
        self.assertRaises(Exception, self.example_func, fail=True)
        errorFound = False
        req_obj = self.client_stub.record_error.last_args['request']
        for row in req_obj.traceback:
            if row.function_name == "example_func":
                errorFound = True
                self.assertEquals('7', row.frame_locals['myvar'])
        self.assertTrue(errorFound)
        self.assertEqual(None, req_obj.error_threshold)


class FuncThreadStub(object):
    def __init__(self, target):
        self.target = target

    def start(self):
        self.target()


class TransportStub(object):
    def open(self):
        pass

    def close(self):
        pass

    def isOpen(self):
        return True


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c) 2011-2013, Shopkick Inc.
# All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# ---
# Author: John Egan <jwegan@gmail.com>

import os
import os.path
import re
from setuptools import setup, find_packages

def markdown_to_reST(text):
  '''This is not a general purpose converter. Only converts this readme'''
  # Convert parameters to italics and prepend a newline
  text = re.sub(pattern=r"\n       (\w+) - (.+)\n",
                repl=r"\n\n       *\g<1>* - \g<2>\n",
                string=text)

  # Parse [http://url](text), and just leave the url
  text = re.sub(pattern=r"\[([^\]]+)\]\([^)]+\)",
                repl=r"\g<1>",
                string=text)

  # Disable formatting of numbered lists
  text = re.sub(pattern=r"\n(\d+). ",
                repl=r"\n\\\g<1>. ",
                string=text)
  return text

setup(
  name='flawless',
  version='0.2.9',
  description='Python Error Monitoring and Reporting',
  long_description=markdown_to_reST(open("README.md").read()),
  license='MPL 2.0',
  author='John Egan',
  author_email='jwegan@gmail.com',
  url='http://github.com/shopkick/flawless',
  zip_safe=False,
  packages=find_packages(exclude=['ez_setup', 'examples', 'packages', 'tests*']),
  data_files=[('config', [os.path.join('config', f) for f in os.listdir('config')
                          if not f.startswith('.')])],
  entry_points="""
  [console_scripts]
  flawless = flawless.server.manage:main
  """,
  classifiers=[
    "Development Status :: 3 - Alpha",
    "Framework :: Pylons",
    "Framework :: Django",
    "Framework :: Pyramid",
    "Framework :: Paste",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
    "Topic :: Software Development :: Bug Tracking",
    "Topic :: Software Development :: Quality Assurance",
  ],
)

